// Copyright 2023 The Forgotten Server Authors. All rights reserved.
// Use of this source code is governed by the GPL-2.0 License that can be found in the LICENSE file.

#ifndef FS_CREATUREEVENT_H
#define FS_CREATUREEVENT_H

#include "baseevents.h"
#include "enums.h"
#include "luascript.h"

class CreatureEvent;
using CreatureEvent_ptr = std::unique_ptr<CreatureEvent>;

enum CreatureEventType_t
{
	CREATURE_EVENT_NONE,
	CREATURE_EVENT_LOGIN,
	CREATURE_EVENT_LOGOUT,
	CREATURE_EVENT_THINK,
	CREATURE_EVENT_PREPAREDEATH,
	CREATURE_EVENT_DEATH,
	CREATURE_EVENT_KILL,
	CREATURE_EVENT_ADVANCE,
	CREATURE_EVENT_TEXTEDIT,
	CREATURE_EVENT_HEALTHCHANGE,
	CREATURE_EVENT_MANACHANGE,
	CREATURE_EVENT_EXTENDED_OPCODE, // otclient additional network opcodes
};

class CreatureEvent final : public Event
{
public:
	explicit CreatureEvent(LuaScriptInterface* interface);

	bool configureEvent(const pugi::xml_node& node) override;

	CreatureEventType_t getEventType() const { return type; }
	void setEventType(CreatureEventType_t eventType) { type = eventType; }
	const std::string& getName() const { return eventName; }
	void setName(std::string_view name) { eventName = name; }
	bool isLoaded() const { return loaded; }
	void setLoaded(bool b) { loaded = b; }

	void clearEvent();
	void copyEvent(CreatureEvent* creatureEvent);

	// scripting
	bool executeOnLogin(Player* player) const;
	bool executeOnLogout(Player* player) const;
	bool executeOnThink(Creature* creature, uint32_t interval);
	bool executeOnPrepareDeath(Creature* creature, Creature* killer);
	bool executeOnDeath(Creature* creature, Item* corpse, Creature* killer, Creature* mostDamageKiller,
	                    bool lastHitUnjustified, bool mostDamageUnjustified);
	void executeOnKill(Creature* creature, Creature* target);
	bool executeAdvance(Player* player, skills_t, uint32_t, uint32_t);
	bool executeTextEdit(Player* player, Item* item, std::string_view text, const uint32_t windowTextId);
	void executeHealthChange(Creature* creature, Creature* attacker, CombatDamage& damage);
	void executeManaChange(Creature* creature, Creature* attacker, CombatDamage& damage);
	void executeExtendedOpcode(Player* player, uint8_t opcode, std::string_view buffer);
	//

private:
	std::string_view getScriptEventName() const override;

	std::string eventName;
	CreatureEventType_t type;
	bool loaded;
};

class CreatureEvents final : public BaseEvents
{
public:
	CreatureEvents();

	// non-copyable
	CreatureEvents(const CreatureEvents&) = delete;
	CreatureEvents& operator=(const CreatureEvents&) = delete;

	// global events
	bool playerLogin(Player* player) const;
	bool playerLogout(Player* player) const;
	bool playerAdvance(Player* player, skills_t, uint32_t, uint32_t);

	CreatureEvent* getEventByName(std::string_view name, bool forceLoaded = true);

	bool registerLuaEvent(CreatureEvent* event);
	void clear(bool fromLua) override final;

	void removeInvalidEvents();

private:
	LuaScriptInterface& getScriptInterface() override;
	std::string_view getScriptBaseName() const override;
	Event_ptr getEvent(std::string_view nodeName) override;
	bool registerEvent(Event_ptr event, const pugi::xml_node& node) override;

	// creature events
	using CreatureEventMap = std::map<std::string, CreatureEvent>;
	CreatureEventMap creatureEvents;

	LuaScriptInterface scriptInterface;
};

#endif
